/*****************************************************************************
    mkgloblist.c

    Diese Datei ist Teil des Programmpakets 'ACL-tftp'.
    Alle Rechte bei Andreas G. Schindler
                    Klappacher Str. 2i
                    64285 Darmstadt, Deutschland.

    Dieses Programm ist 'Freie Software'; Sie knnen es entsprechend den 
    Bedingungen der 'Allgemeinen ffentlichen GNU-Lizenz' (verffentlicht 
    durch die 'Free Software Foundation') weiter verbreiten und/oder 
    modifizieren. Massgeblich ist die Version 2 oder eine neuere Version
    dieser Lizenz. 

    Dieses Programm wird in der Hoffnung dass es ntzlich is weitergegeben,
    jedoch OHNE JEDE GARANTIE. Dieser Garantieauschluss bezieht sich auch
    auf Eigenschaften wie Marktfhigkeit oder Eignung zu einem bestimmten
    Zweck. Details finden sich in der 'Allgemeinen ffentlichen GNU-Lizenz'.

    Eine Kopie dieser Lizenz sollte in  Programmpaket, dem diese Datei
    entstammt, enthalten sein. Falls nicht, schreiben Sie an die 
    'Free Software Foundation, Inc.', 59 Temple Place, Suite 330, Boston, 
    MA  02111-1307  USA.


    This file is part of the 'ACL-tftp' Package
    Copyright (C) 1999-2003 by Andreas G. Schindler
                               Klappacher Str. 2i
                               Darmstadt, Germany.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include "mf.h"
#include "getarg.h"

char *warn = 
       "\n/* ----- %s of generated section - do not edit ----- */\n\n";

char *plist[] = { "MF_NONE", "MF_PRESET", "MF_DEFAULT", "MF_CMD",
                  "MF_CONFIG", "MF_RUN", NULL };

/*
        remove nonprintable characters from end of a string
*/

void trim (char *txt)
{
    char *p;

    for (p = txt+strlen(txt)-1; p >= txt; p--) {
        if (isgraph((int)*p))
            break;
        *p = '\0';
    }
}

void repl(char *txt)
{
   char *p;

   for (p = txt; *p != '\0'; p++)
       if (*p == '~') *p = ' ';
   trim(txt);
}

void clean(char **txt)
{
   char *q, *p = *txt;
   if (*p == '\0')
       return;
   for (; *p != '\0'; p++) {
       if (strchr(" \t/*", *p) == NULL) break;
   }
   *txt = p;
   for (q=(p+strlen(p)-1);; q--) {
       if (strchr(" \t/*", *q) == NULL) break;
       *q = '\0';
   }
}
   
void exch(char **txt)
{
   char *q, *p = *txt;

   switch (*p) {
       case 'C':
           q = "MF_CMD";
           break;
       case 'G':
           q = "(MF_CMD|MF_CONFIG)";
           break;
       case 'F':
           q = "MF_CONFIG";
           break;
       case 'R':
           q = "MF_RUN";
           break;
       case 'I':
           q = "MF_NONE";
           break;
       default:
           q = "";
           break;
   }
   *txt = q;
}

int empty(char *txt)
{
    for(;*txt != '\0'; txt++)
        if (*txt > ' ') return 0;
    return 1;
}

int incom(char *txt)
{
    static int com=0, prev;

    for(prev=com; *txt != '\0'; txt++) {
        if ((*txt == '/') && (*(txt+1) == '*')) {
            txt++;
            com=prev=1;
            continue;
        } 
        if ((*txt == '*') && (*(txt+1) == '/')) {
            txt++;
            com=0;
            continue;
        } 
    }
    return prev;
}

char *sem(char *txt)
{
    static char buf[1024];

    strcpy(buf, txt);
    strcat(buf, ";");
    return buf;
}


int isarray(char *txt)
{
    return (strchr(txt,'[') != NULL);
}


char *nonarray(char *txt)
{
    static char buf[1024];
    char *p;

    strcpy(buf, txt);
    if ((p=strchr(buf,'[')) != NULL)
        *p = 0;
    return buf;
}


int main (int argc, char **argv)
{
    struct glob 
    { char *typ; char *nam; char *set; char *ini; char *com; int f;} 
        glist[1024];
    char line[1024];
    char *tok;
    int i, n=0, k, j, cfg=0;
    char *allopts = "hf?";

    if (getarg(argc,argv,allopts,'?',"",NULL) != NULL) {
        char *basename;
        basename = strrchr(argv[0], '/');
        if (basename == NULL)
            basename = argv[0];
        else
            basename++;
        fprintf (stderr,"Usage is: %s {-h|-f} [-?]\n"
                        "\t-h\tbuild header section\n"
                        "\t-f\tbuild function section\n\n",basename);
        return 1;
    }

    for (n=0; fgets(line,sizeof(line),stdin) != NULL; n++) {
        line[strlen(line)-1] = '\0';
        if (empty(line) || incom(line)) {
            glist[n].f = 1;
            glist[n].typ = strdup(line);
            continue;
        }
        else
            glist[n].f = 0;
        for (i = 0; i < (int)(sizeof(glist)/sizeof(glist[0])); i++) {
            switch (i) {
                case 0:
                    tok = strtok(line," \t");
                    glist[n].typ = (tok==NULL) ? "" : strdup(tok);
                    repl (glist[n].typ);
                    break;
                case 1:
                    tok = strtok(NULL," \t");
                    glist[n].nam = (tok==NULL) ? "" : strdup(tok);
                    repl (glist[n].nam);
                    break;
                case 2:
                    tok = strtok(NULL," \t");
                    glist[n].set = (tok==NULL) ? "" : strdup(tok);
                    repl (glist[n].set);
                    exch (&glist[n].set);
                    break;
                case 3:
                    tok = strtok(NULL," \t");
                    glist[n].ini = (tok==NULL) ? "" : strdup(tok);
                    repl (glist[n].ini);
                    break;
                case 4:
                    tok = strtok(NULL,"\0");
                    glist[n].com = (tok==NULL) ? "" : strdup(tok);
                    clean (&glist[n].com);
                    break;
            }
        }
    }

    
    if (getarg(argc,argv,allopts,'h',"",NULL) != NULL) {
        printf (warn, "start");
        printf ("#ifndef GLOBDEF\n\n");
        for (i=0; i<n; i++) {
            if (glist[i].f) {
                printf("%s\n", glist[i].typ);
                continue;
            }
            printf ("extern %-12s%-17s /* %-30s */\n", 
	    	    glist[i].typ, sem(glist[i].nam), glist[i].com);
        }
        printf ("\n");
        for (i=0; i<n; i++) {
            if (glist[i].f) continue;
            if (!strlen(glist[i].set)) continue;
            printf ("extern %-12sMF_%s;\n", "int" , nonarray(glist[i].nam)); 
        }
        printf ("\n#else /* GLOBDEF */\n\n");
        for (i=0; i<n; i++) {
            if (glist[i].f) continue;
            if (strlen(glist[i].ini))
                printf ("%-12s%-16s = %s;\n", 
	    	    glist[i].typ, glist[i].nam, glist[i].ini);
            else
                printf ("%-12s%s;\n", 
	    	    glist[i].typ, glist[i].nam);
        }
        printf ("\n");
        for (i=0; i<n; i++) {
            if (glist[i].f) continue;
            if (!strlen(glist[i].set)) continue;
            printf ("%-12sMF_%-13s = MF_%s;\n", "int" , 
                    nonarray(glist[i].nam), 
                    strcmp(glist[i].set, "MF_NONE") ? "PRESET" : "NONE");
        }
        printf ("\n#endif /* GLOBDEF */\n\n");
        printf ("\n#ifdef GLOBRESET\n\n");
        for (i=0; i<n; i++) {
            if (glist[i].f) continue;
            if (!strlen(glist[i].ini) || !strlen(glist[i].set)) continue;
            printf ("static %-12sI_%-16s = %s;\n", 
	    	    glist[i].typ, glist[i].nam, glist[i].ini);
        }
        printf ("\n#endif /* GLOBRESET */\n\n");
        printf ("void globReset(int flag);\n\n");
        printf (warn, "end");
    }
    if (getarg(argc,argv,allopts,'f',"",NULL) != NULL) {
        printf (warn, "start");
        printf ("#define GLOBRESET\n"
                "#include \"globals.h\"\n\n"
                "void globReset(int flag)\n{\n");
        for (k=0, j=0; plist[k]!=NULL; k++, j=0) {
            cfg = !strcmp(plist[k], "MF_CONFIG");
            for (i=0; i<n; i++) {
                if (glist[i].f) continue;
                if (strstr(glist[i].set, plist[k]) == NULL) continue;
                if (!strlen(glist[i].ini) || !strlen(glist[i].set)) continue;
                if (!j++)
                    printf ("    if (flag & %s) {\n", plist[k]);
                if (cfg) {
                    if (isarray(glist[i].nam)) 
                        printf ("        if (!(MF_%s & MF_CMD))\n"
                                "            strcpy(%s, I_%s);\n",
                                nonarray(glist[i].nam), 
                                nonarray(glist[i].nam), nonarray(glist[i].nam));
                    else
                        printf ("        if (!(MF_%s & MF_CMD))\n"
                                "            %-16s = I_%s;\n", 
                                glist[i].nam, glist[i].nam, glist[i].nam);
                }
                else {
                    if (isarray(glist[i].nam)) 
                        printf ("        strcpy(%s, I_%s);\n",
                                nonarray(glist[i].nam), nonarray(glist[i].nam));
                    else
                        printf ("        %-16s = I_%s;\n", 
                                glist[i].nam, glist[i].nam);
                }
            }
            if (j)
                printf ("    }\n");
        }    
        printf ("}\n");
        printf (warn, "end");
    }
    return 0;
}
